package metrics

import (
	"os"
	"time"

	"github.com/gin-gonic/gin"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	"github.com/spf13/cast"
)

var (
	// namespace = "xmall"
	namespace = ""
	// uptime = prometheus.NewCounterVec(
	// 	prometheus.CounterOpts{
	// 		Namespace: namespace,
	// 		Name:      "uptime",
	// 		Help:      "HTTP service uptime, updated every minute",
	// 	}, nil,
	// )

	reqCount = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Namespace: namespace,
			Name:      "client_code_total",
			Help:      "Total number of HTTP requests made.",
		}, []string{"kind", "operation", "code"},
	)

	reqDuration = prometheus.NewHistogramVec(prometheus.HistogramOpts{
		Namespace: namespace,
		Name:      "server_requests_duration_sec",
		Help:      "Server requests duratio(sec).",
		Buckets:   []float64{0.005, 0.01, 0.025, 0.05, 0.1, 0.250, 0.5, 1},
	}, []string{"kind", "operation"})
)

func init() {
	if ns := os.Getenv("CONTAINER_NAMESPACE"); ns != "" {
		namespace = ns
	}
}

// init registers the prometheus metrics
func initPrometheus() {
	// prometheus.MustRegister(uptime, reqCount, reqDuration)
	prometheus.MustRegister(reqCount, reqDuration)
}

// recordUptime increases service uptime per 1 minute.
// func recordUptime() {
// 	for range time.Tick(time.Minute) {
// 		uptime.WithLabelValues().Inc()
// 	}
// }

// metricsHandler wrappers the standard http.Handler to gin.HandlerFunc
func metricsHandler() gin.HandlerFunc {
	return func(c *gin.Context) {
		handler := promhttp.Handler()
		handler.ServeHTTP(c.Writer, c.Request)
	}
}

// Metrics returns a gin.HandlerFunc for exporting some Web metrics
func Metrics(r *gin.Engine, opts ...Option) gin.HandlerFunc {
	o := defaultOptions()
	o.apply(opts...)

	// init prometheus
	initPrometheus()

	r.GET(o.metricsPath, metricsHandler())

	return func(c *gin.Context) {
		start := time.Now()
		c.Next()

		ok := o.isIgnoreCodeStatus(c.Writer.Status()) ||
			o.isIgnorePath(c.Request.URL.Path) ||
			o.checkIgnoreMethod(c.Request.Method)
		if ok {
			return
		}

		reqCount.With(prometheus.Labels{
			"kind":      "http",
			"operation": c.Request.Method + "." + c.Request.URL.Path,
			"code":      cast.ToString(c.Writer.Status()),
		}).Inc()

		reqDuration.With(prometheus.Labels{
			"kind":      "http",
			"operation": c.Request.Method + "." + c.Request.URL.Path,
		}).Observe(time.Since(start).Seconds())
	}
}
