package aliyun

import (
	"io"
	"mime/multipart"

	"gitlab.wanzhuangkj.com/tush/xpkg/xerrors/xerror"

	"github.com/aliyun/aliyun-oss-go-sdk/oss"
)

type AliyunOSS struct{}

var OSS = &AliyunOSS{}

type AliOssConfig struct {
	Endpoint        string `mapstructure:"endpoint" json:"endpoint" yaml:"endpoint"`
	AccessKeyId     string `mapstructure:"access-key-id" json:"access-key-id" yaml:"access-key-id"`
	AccessKeySecret string `mapstructure:"access-key-secret" json:"access-key-secret" yaml:"access-key-secret"`
	BucketName      string `mapstructure:"bucket-name" json:"bucket-name" yaml:"bucket-name"`
	BasePath        string `mapstructure:"base-path" json:"base-path" yaml:"base-path"`
	BucketUrl       string `mapstructure:"bucket-url" json:"bucket-url" yaml:"bucket-url"`
	Region          string `mapstructure:"region" json:"region" yaml:"region"`
}

func NewClient(cfg *AliOssConfig) (bucket *oss.Bucket, err error) {
	client, err := oss.New(cfg.Endpoint,
		cfg.AccessKeyId,
		cfg.AccessKeySecret,
		oss.AuthVersion(oss.AuthV4),
		oss.Region(cfg.Region))
	if err != nil {
		return nil, xerror.New(err.Error())
	}
	bucket, err = client.Bucket(cfg.BucketName)
	return
}

func NewBucket(cfg *AliOssConfig) (bucket *oss.Bucket, err error) {
	client, err := oss.New(cfg.Endpoint, cfg.AccessKeyId, cfg.AccessKeySecret)
	if err != nil {
		return
	}
	bucket, err = client.Bucket(cfg.BucketName)
	return
}

func (*AliyunOSS) UploadFile(reader io.Reader, name string, cfg *AliOssConfig) (string, string, error) {
	bucket, err := NewClient(cfg)
	if err != nil {
		return "", "", err
	}
	fileTmpPath := cfg.BasePath + "/" + name
	err = bucket.PutObject(fileTmpPath, reader)
	if err != nil {
		return "", "", xerror.New(err.Error())
	}
	return cfg.BucketUrl + "/" + fileTmpPath, fileTmpPath, nil
}

type Attachment struct {
	File        io.Reader
	Name        string
	AbsUrl      string
	RelativeUrl string
}

func (x Attachment) NewFromMultipartFileHeaders(files []*multipart.FileHeader) ([]*Attachment, error) {
	if len(files) == 0 {
		return nil, nil
	}
	attachments := make([]*Attachment, len(files))
	for i := range files {
		reader, err := files[i].Open()
		if err != nil {
			return nil, xerror.New(err.Error())
		}
		attachment := &Attachment{
			File: reader,
			Name: files[i].Filename,
		}
		attachments = append(attachments, attachment)
	}
	return attachments, nil
}

func (x *AliyunOSS) UploadMultipartFileHeaders(files []*multipart.FileHeader, cfg *AliOssConfig) ([]*Attachment, error) {
	if len(files) == 0 {
		return nil, nil
	}
	attachments, err := Attachment{}.NewFromMultipartFileHeaders(files)
	if err != nil {
		return nil, err
	}
	err = x.UploadFiles(&attachments, cfg)
	if err != nil {
		return nil, err
	}

	return attachments, nil
}

func (*AliyunOSS) UploadFiles(files *[]*Attachment, cfg *AliOssConfig) error {
	if len(*files) == 0 {
		return nil
	}
	bucket, err := NewClient(cfg)
	if err != nil {
		return xerror.New(err.Error())
	}
	for i, file := range *files {
		relativeUri := cfg.BasePath + "/" + file.Name
		err = bucket.PutObject(relativeUri, file.File)
		if err != nil {
			return xerror.New(err.Error())
		}
		(*files)[i].AbsUrl = cfg.BucketUrl + "/" + relativeUri
		(*files)[i].RelativeUrl = relativeUri
	}
	return nil
}
